=== modified file 'NEWS'
--- NEWS	2009-09-07 23:47:14 +0000
+++ NEWS	2009-09-08 06:28:13 +0000
@@ -76,9 +76,9 @@
   single stream, reducing layering churn. (Robert Collins)
 
 * Prevent some kinds of incomplete data from being committed to a 2a
-  repository, such as revisions without inventories or inventories without
-  chk_bytes root records.
-  (Andrew Bennetts, #423506)
+  repository, such as revisions without inventories, a missing chk_bytes
+  record for an inventory, or a missing text referenced by an inventory.
+  (Andrew Bennetts, #423506, #406687)
   
 Documentation
 *************

=== modified file 'bzrlib/groupcompress.py'
--- bzrlib/groupcompress.py	2009-09-04 07:39:08 +0000
+++ bzrlib/groupcompress.py	2009-09-08 06:25:26 +0000
@@ -1168,20 +1168,28 @@
 class GroupCompressVersionedFiles(VersionedFiles):
     """A group-compress based VersionedFiles implementation."""
 
-    def __init__(self, index, access, delta=True):
+    def __init__(self, index, access, delta=True, _unadded_refs=None):
         """Create a GroupCompressVersionedFiles object.
 
         :param index: The index object storing access and graph data.
         :param access: The access object storing raw data.
         :param delta: Whether to delta compress or just entropy compress.
+        :param _unadded_refs: private parameter, don't use.
         """
         self._index = index
         self._access = access
         self._delta = delta
-        self._unadded_refs = {}
+        if _unadded_refs is None:
+            _unadded_refs = {}
+        self._unadded_refs = _unadded_refs
         self._group_cache = LRUSizeCache(max_size=50*1024*1024)
         self._fallback_vfs = []
 
+    def without_fallbacks(self):
+        """Return a clone of this object without any fallbacks configured."""
+        return GroupCompressVersionedFiles(self._index, self._access,
+            self._delta, _unadded_refs=dict(self._unadded_refs))
+
     def add_lines(self, key, parents, lines, parent_texts=None,
         left_matching_blocks=None, nostore_sha=None, random_id=False,
         check_content=True):

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2009-08-27 05:22:14 +0000
+++ bzrlib/remote.py	2009-09-08 08:10:19 +0000
@@ -1739,6 +1739,7 @@
             # The stream included an inventory-delta record, but the remote
             # side isn't new enough to support them.  So we need to send the
             # rest of the stream via VFS.
+            self.target_repo.refresh_data()
             return self._resume_stream_with_vfs(response, src_format)
         if response[0][0] == 'missing-basis':
             tokens, missing_keys = bencode.bdecode_as_tuple(response[0][1])

=== modified file 'bzrlib/repofmt/groupcompress_repo.py'
--- bzrlib/repofmt/groupcompress_repo.py	2009-09-07 03:00:23 +0000
+++ bzrlib/repofmt/groupcompress_repo.py	2009-09-08 05:51:36 +0000
@@ -588,21 +588,21 @@
         """Detect missing inventories or chk root entries for the new revisions
         in this write group.
 
-        :returns: set of missing keys.  Note that not every missing key is
-            guaranteed to be reported.
+        :returns: list of strs, summarising any problems found.  If the list is
+            empty no problems were found.
         """
-        if getattr(self.repo, 'chk_bytes', None) is None:
-            return set()
         # Ensure that all revisions added in this write group have:
         #   - corresponding inventories,
         #   - chk root entries for those inventories,
         #   - and any present parent inventories have their chk root
         #     entries too.
         # And all this should be independent of any fallback repository.
+        problems = []
         key_deps = self.repo.revisions._index._key_dependencies
         new_revisions_keys = key_deps.get_new_keys()
         no_fallback_inv_index = self.repo.inventories._index
         no_fallback_chk_bytes_index = self.repo.chk_bytes._index
+        no_fallback_texts_index = self.repo.texts._index
         inv_parent_map = no_fallback_inv_index.get_parent_map(
             new_revisions_keys)
         # Are any inventories for corresponding to the new revisions missing?
@@ -610,7 +610,9 @@
         missing_corresponding = set(new_revisions_keys)
         missing_corresponding.difference_update(corresponding_invs)
         if missing_corresponding:
-            return [('inventories', key) for key in missing_corresponding]
+            problems.append("inventories missing for revisions %s" %
+                (sorted(missing_corresponding),))
+            return problems
         # Are any chk root entries missing for any inventories?  This includes
         # any present parent inventories, which may be used when calculating
         # deltas for streaming.
@@ -620,17 +622,54 @@
         # Filter out ghost parents.
         all_inv_keys.intersection_update(
             no_fallback_inv_index.get_parent_map(all_inv_keys))
+        parent_invs_only_keys = all_inv_keys.symmetric_difference(
+            corresponding_invs)
         all_missing = set()
         inv_ids = [key[-1] for key in all_inv_keys]
-        for inv in self.repo.iter_inventories(inv_ids, 'unordered'):
-            root_keys = set([inv.id_to_entry.key()])
-            if inv.parent_id_basename_to_file_id is not None:
-                root_keys.add(inv.parent_id_basename_to_file_id.key())
-            present = no_fallback_chk_bytes_index.get_parent_map(root_keys)
-            missing = root_keys.difference(present)
-            all_missing.update([('chk_bytes',) + key for key in missing])
-        return all_missing
-        
+        parent_invs_only_ids = [key[-1] for key in parent_invs_only_keys]
+        root_key_info = _build_interesting_key_sets(
+            self.repo, inv_ids, parent_invs_only_ids)
+        expected_chk_roots = root_key_info.all_keys()
+        present_chk_roots = no_fallback_chk_bytes_index.get_parent_map(
+            expected_chk_roots)
+        missing_chk_roots = expected_chk_roots.difference(present_chk_roots)
+        if missing_chk_roots:
+            problems.append("missing referenced chk root keys: %s"
+                % (sorted(missing_chk_roots),))
+            # Don't bother checking any further.
+            return problems
+        # Find all interesting chk_bytes records, and make sure they are
+        # present, as well as the text keys they reference.
+        chk_bytes_no_fallbacks = self.repo.chk_bytes.without_fallbacks()
+        chk_bytes_no_fallbacks._search_key_func = \
+            self.repo.chk_bytes._search_key_func
+        chk_diff = chk_map.iter_interesting_nodes(
+            chk_bytes_no_fallbacks, root_key_info.interesting_root_keys,
+            root_key_info.uninteresting_root_keys)
+        bytes_to_info = inventory.CHKInventory._bytes_to_utf8name_key
+        text_keys = set()
+        try:
+            for record in _filter_text_keys(chk_diff, text_keys, bytes_to_info):
+                pass
+        except errors.NoSuchRevision, e:
+            # XXX: It would be nice if we could give a more precise error here.
+            problems.append("missing chk node(s) for id_to_entry maps")
+        chk_diff = chk_map.iter_interesting_nodes(
+            chk_bytes_no_fallbacks, root_key_info.interesting_pid_root_keys,
+            root_key_info.uninteresting_pid_root_keys)
+        try:
+            for interesting_rec, interesting_map in chk_diff:
+                pass
+        except errors.NoSuchRevision, e:
+            problems.append(
+                "missing chk node(s) for parent_id_basename_to_file_id maps")
+        present_text_keys = no_fallback_texts_index.get_parent_map(text_keys)
+        missing_text_keys = text_keys.difference(present_text_keys)
+        if missing_text_keys:
+            problems.append("missing text keys: %r"
+                % (sorted(missing_text_keys),))
+        return problems
+
     def _execute_pack_operations(self, pack_operations,
                                  _packer_class=GCCHKPacker,
                                  reload_func=None):
@@ -898,17 +937,12 @@
                                         parent_keys)
             present_parent_inv_ids = set(
                 [k[-1] for k in present_parent_inv_keys])
-            uninteresting_root_keys = set()
-            interesting_root_keys = set()
             inventories_to_read = set(revision_ids)
             inventories_to_read.update(present_parent_inv_ids)
-            for inv in self.iter_inventories(inventories_to_read):
-                entry_chk_root_key = inv.id_to_entry.key()
-                if inv.revision_id in present_parent_inv_ids:
-                    uninteresting_root_keys.add(entry_chk_root_key)
-                else:
-                    interesting_root_keys.add(entry_chk_root_key)
-
+            root_key_info = _build_interesting_key_sets(
+                self, inventories_to_read, present_parent_inv_ids)
+            interesting_root_keys = root_key_info.interesting_root_keys
+            uninteresting_root_keys = root_key_info.uninteresting_root_keys
             chk_bytes = self.chk_bytes
             for record, items in chk_map.iter_interesting_nodes(chk_bytes,
                         interesting_root_keys, uninteresting_root_keys,
@@ -1048,13 +1082,10 @@
         bytes_to_info = inventory.CHKInventory._bytes_to_utf8name_key
         chk_bytes = self.from_repository.chk_bytes
         def _filter_id_to_entry():
-            for record, items in chk_map.iter_interesting_nodes(chk_bytes,
-                        self._chk_id_roots, uninteresting_root_keys):
-                for name, bytes in items:
-                    # Note: we don't care about name_utf8, because we are always
-                    # rich-root = True
-                    _, file_id, revision_id = bytes_to_info(bytes)
-                    self._text_keys.add((file_id, revision_id))
+            interesting_nodes = chk_map.iter_interesting_nodes(chk_bytes,
+                        self._chk_id_roots, uninteresting_root_keys)
+            for record in _filter_text_keys(interesting_nodes, self._text_keys,
+                    bytes_to_info):
                 if record is not None:
                     yield record
             # Consumed
@@ -1098,7 +1129,7 @@
             missing_inventory_keys.add(key[1:])
         if self._chk_id_roots or self._chk_p_id_roots:
             raise AssertionError('Cannot call get_stream_for_missing_keys'
-                ' untill all of get_stream() has been consumed.')
+                ' until all of get_stream() has been consumed.')
         # Yield the inventory stream, so we can find the chk stream
         # Some of the missing_keys will be missing because they are ghosts.
         # As such, we can ignore them. The Sink is required to verify there are
@@ -1111,6 +1142,54 @@
             yield stream_info
 
 
+class _InterestingKeyInfo(object):
+    def __init__(self):
+        self.interesting_root_keys = set()
+        self.interesting_pid_root_keys = set()
+        self.uninteresting_root_keys = set()
+        self.uninteresting_pid_root_keys = set()
+
+    def all_interesting(self):
+        return self.interesting_root_keys.union(self.interesting_pid_root_keys)
+
+    def all_uninteresting(self):
+        return self.uninteresting_root_keys.union(
+            self.uninteresting_pid_root_keys)
+
+    def all_keys(self):
+        return self.all_interesting().union(self.all_uninteresting())
+
+
+def _build_interesting_key_sets(repo, inventory_ids, parent_only_inv_ids):
+    result = _InterestingKeyInfo()
+    for inv in repo.iter_inventories(inventory_ids, 'unordered'):
+        root_key = inv.id_to_entry.key()
+        pid_root_key = inv.parent_id_basename_to_file_id.key()
+        if inv.revision_id in parent_only_inv_ids:
+            result.uninteresting_root_keys.add(root_key)
+            result.uninteresting_pid_root_keys.add(pid_root_key)
+        else:
+            result.interesting_root_keys.add(root_key)
+            result.interesting_pid_root_keys.add(pid_root_key)
+    return result
+
+
+def _filter_text_keys(interesting_nodes_iterable, text_keys, bytes_to_info):
+    """Iterate the result of iter_interesting_nodes, yielding the records
+    and adding to text_keys.
+    """
+    for record, items in interesting_nodes_iterable:
+        for name, bytes in items:
+            # Note: we don't care about name_utf8, because groupcompress repos
+            # are always rich-root, so there are no synthesised root records to
+            # ignore.
+            _, file_id, revision_id = bytes_to_info(bytes)
+            text_keys.add((file_id, revision_id))
+        yield record
+
+
+
+
 class RepositoryFormatCHK1(RepositoryFormatPack):
     """A hashed CHK+group compress pack repository."""
 

=== modified file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	2009-09-07 03:00:23 +0000
+++ bzrlib/repofmt/pack_repo.py	2009-09-08 05:51:36 +0000
@@ -2066,12 +2066,12 @@
     def _check_new_inventories(self):
         """Detect missing inventories in this write group.
 
-        :returns: set of missing keys.  Note that not every missing key is
-            guaranteed to be reported.
+        :returns: list of strs, summarising any problems found.  If the list is
+            empty no problems were found.
         """
         # The base implementation does no checks.  GCRepositoryPackCollection
         # overrides this.
-        return set()
+        return []
         
     def _commit_write_group(self):
         all_missing = set()
@@ -2087,11 +2087,11 @@
             raise errors.BzrCheckError(
                 "Repository %s has missing compression parent(s) %r "
                  % (self.repo, sorted(all_missing)))
-        all_missing = self._check_new_inventories()
-        if all_missing:
+        problems = self._check_new_inventories()
+        if problems:
+            problems_summary = '\n'.join(problems)
             raise errors.BzrCheckError(
-                "Repository %s missing keys for new revisions %r "
-                 % (self.repo, sorted(all_missing)))
+                "Cannot add revision(s) to repository: " + problems_summary)
         self._remove_pack_indices(self._new_pack)
         any_new_content = False
         if self._new_pack.data_inserted():

=== modified file 'bzrlib/tests/per_interrepository/test_fetch.py'
--- bzrlib/tests/per_interrepository/test_fetch.py	2009-08-26 07:40:38 +0000
+++ bzrlib/tests/per_interrepository/test_fetch.py	2009-09-09 02:14:12 +0000
@@ -45,8 +45,16 @@
     )
 
 
+
 class TestInterRepository(TestCaseWithInterRepository):
 
+    def disable_commit_write_group_paranoia(self, repo):
+        pack_coll = getattr(repo, '_pack_collection', None)
+        if pack_coll is not None:
+            # Monkey-patch the pack collection instance to allow storing
+            # incomplete revisions.
+            pack_coll._check_new_inventories = lambda: []
+
     def test_fetch(self):
         tree_a = self.make_branch_and_tree('a')
         self.build_tree(['a/foo'])
@@ -365,11 +373,16 @@
         to_repo.lock_write()
         try:
             to_repo.start_write_group()
-            inv = tree.branch.repository.get_inventory('rev-one')
-            to_repo.add_inventory('rev-one', inv, [])
-            rev = tree.branch.repository.get_revision('rev-one')
-            to_repo.add_revision('rev-one', rev, inv=inv)
-            to_repo.commit_write_group()
+            try:
+                inv = tree.branch.repository.get_inventory('rev-one')
+                to_repo.add_inventory('rev-one', inv, [])
+                rev = tree.branch.repository.get_revision('rev-one')
+                to_repo.add_revision('rev-one', rev, inv=inv)
+                self.disable_commit_write_group_paranoia(to_repo)
+                to_repo.commit_write_group()
+            except:
+                to_repo.abort_write_group(suppress_errors=True)
+                raise
         finally:
             to_repo.unlock()
 
@@ -424,7 +437,7 @@
         source_tree.add(['id'], ['id'])
         source_tree.commit('a', rev_id='a')
         # now we manually insert a revision with an inventory referencing
-        # 'id' at revision 'b', but we do not insert revision b.
+        # file 'id' at revision 'b', but we do not insert revision b.
         # this should ensure that the new versions of files are being checked
         # for during pull operations
         inv = source.get_inventory('a')
@@ -442,6 +455,7 @@
                        revision_id='b')
         rev.parent_ids = ['a']
         source.add_revision('b', rev)
+        self.disable_commit_write_group_paranoia(source)
         source.commit_write_group()
         self.assertRaises(errors.RevisionNotPresent, target.fetch, source)
         self.assertFalse(target.has_revision('b'))

=== modified file 'bzrlib/tests/per_repository/test_fileid_involved.py'
--- bzrlib/tests/per_repository/test_fileid_involved.py	2009-06-10 03:56:49 +0000
+++ bzrlib/tests/per_repository/test_fileid_involved.py	2009-09-09 02:14:12 +0000
@@ -19,6 +19,8 @@
 import time
 
 from bzrlib import (
+    errors,
+    remote,
     revision as _mod_revision,
     tests,
     )
@@ -342,9 +344,23 @@
         self.addCleanup(b.unlock)
         b.repository.start_write_group()
         b.repository.add_revision('B-id', new_rev, new_inv)
+        self.disable_commit_write_group_paranoia(b.repository)
         b.repository.commit_write_group()
         return b
 
+    def disable_commit_write_group_paranoia(self, repo):
+        if isinstance(repo, remote.RemoteRepository):
+            # We can't easily disable the checks in a remote repo.
+            repo.abort_write_group()
+            raise TestSkipped(
+                "repository format does not support storing revisions with "
+                "missing texts.")
+        pack_coll = getattr(repo, '_pack_collection', None)
+        if pack_coll is not None:
+            # Monkey-patch the pack collection instance to allow storing
+            # incomplete revisions.
+            pack_coll._check_new_inventories = lambda: []
+
     def test_file_ids_include_ghosts(self):
         b = self.create_branch_with_ghost_text()
         repo = b.repository

=== modified file 'bzrlib/tests/per_repository/test_reconcile.py'
--- bzrlib/tests/per_repository/test_reconcile.py	2009-06-10 03:56:49 +0000
+++ bzrlib/tests/per_repository/test_reconcile.py	2009-09-08 11:32:53 +0000
@@ -341,6 +341,9 @@
         repo.start_write_group()
         inv = Inventory(revision_id='wrong-first-parent')
         inv.root.revision = 'wrong-first-parent'
+        if repo.supports_rich_root():
+            root_id = inv.root.file_id
+            repo.texts.add_lines((root_id, 'wrong-first-parent'), [], [])
         sha1 = repo.add_inventory('wrong-first-parent', inv, ['2', '1'])
         rev = Revision(timestamp=0,
                        timezone=None,

=== modified file 'bzrlib/tests/per_repository/test_repository.py'
--- bzrlib/tests/per_repository/test_repository.py	2009-08-18 22:03:18 +0000
+++ bzrlib/tests/per_repository/test_repository.py	2009-09-08 11:32:53 +0000
@@ -760,8 +760,8 @@
         repo = self.make_repository('repo')
         repo.lock_write()
         self.addCleanup(repo.unlock)
-        self.addCleanup(repo.commit_write_group)
         repo.start_write_group()
+        self.addCleanup(repo.abort_write_group)
         inv = Inventory(revision_id='A')
         inv.root.revision = 'A'
         repo.add_inventory('A', inv, [])
@@ -788,6 +788,8 @@
         repo = self.make_repository('repo')
         repo.lock_write()
         repo.start_write_group()
+        root_id = inv.root.file_id
+        repo.texts.add_lines(('fixed-root', 'A'), [], [])
         repo.add_revision('A', Revision('A', committer='B', timestamp=0,
                           timezone=0, message='C'), inv=inv)
         repo.commit_write_group()
@@ -1225,6 +1227,9 @@
         repo.start_write_group()
         inv = Inventory(revision_id = 'ghost')
         inv.root.revision = 'ghost'
+        if repo.supports_rich_root():
+            root_id = inv.root.file_id
+            repo.texts.add_lines((root_id, 'ghost'), [], [])
         sha1 = repo.add_inventory('ghost', inv, [])
         rev = bzrlib.revision.Revision(timestamp=0,
                                        timezone=None,
@@ -1240,6 +1245,9 @@
 
         inv = Inventory(revision_id = 'the_ghost')
         inv.root.revision = 'the_ghost'
+        if repo.supports_rich_root():
+            root_id = inv.root.file_id
+            repo.texts.add_lines((root_id, 'the_ghost'), [], [])
         sha1 = repo.add_inventory('the_ghost', inv, [])
         rev = bzrlib.revision.Revision(timestamp=0,
                                        timezone=None,

=== modified file 'bzrlib/tests/per_repository/test_write_group.py'
--- bzrlib/tests/per_repository/test_write_group.py	2009-09-02 03:07:23 +0000
+++ bzrlib/tests/per_repository/test_write_group.py	2009-09-08 06:25:26 +0000
@@ -28,9 +28,8 @@
     versionedfile,
     )
 from bzrlib.branch import BzrBranchFormat7
-from bzrlib.inventory import InventoryDirectory
 from bzrlib.transport import local, memory
-from bzrlib.tests import KnownFailure, TestNotApplicable
+from bzrlib.tests import TestNotApplicable
 from bzrlib.tests.per_repository import TestCaseWithRepository
 
 
@@ -361,143 +360,6 @@
         sink.insert_stream((), repo._format, tokens)
         self.assertEqual([True], call_log)
 
-    def test_missing_chk_root_for_inventory(self):
-        """commit_write_group fails with BzrCheckError when the chk root record
-        for a new inventory is missing.
-        """
-        builder = self.make_branch_builder('simple-branch')
-        builder.build_snapshot('A-id', None, [
-            ('add', ('', 'root-id', 'directory', None)),
-            ('add', ('file', 'file-id', 'file', 'content\n'))])
-        b = builder.get_branch()
-        if not b.repository._format.supports_chks:
-            raise TestNotApplicable('requires repository with chk_bytes')
-        b.lock_read()
-        self.addCleanup(b.unlock)
-        repo = self.make_repository('damaged-repo')
-        repo.lock_write()
-        repo.start_write_group()
-        # Now, add the objects manually
-        text_keys = [('file-id', 'A-id'), ('root-id', 'A-id')]
-        # Directly add the texts, inventory, and revision object for 'A-id' --
-        # but don't add the chk_bytes.
-        src_repo = b.repository
-        repo.texts.insert_record_stream(src_repo.texts.get_record_stream(
-            text_keys, 'unordered', True))
-        repo.inventories.insert_record_stream(
-            src_repo.inventories.get_record_stream(
-                [('A-id',)], 'unordered', True))
-        repo.revisions.insert_record_stream(
-            src_repo.revisions.get_record_stream(
-                [('A-id',)], 'unordered', True))
-        # Make sure the presence of the missing data in a fallback does not
-        # avoid the error.
-        repo.add_fallback_repository(b.repository)
-        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
-        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
-        self.assertRaises(
-            errors.BzrCheckError, reopened_repo.commit_write_group)
-        reopened_repo.abort_write_group()
-
-    def test_missing_chk_root_for_unchanged_inventory(self):
-        """commit_write_group fails with BzrCheckError when the chk root record
-        for a new inventory is missing, even if the parent inventory is present
-        and has identical content (i.e. the same chk root).
-        
-        A stacked repository containing only a revision with an identical
-        inventory to its parent will still have the chk root records for those
-        inventories.
-
-        (In principle the chk records are unnecessary in this case, but in
-        practice bzr 2.0rc1 (at least) expects to find them.)
-        """
-        # Make a branch where the last two revisions have identical
-        # inventories.
-        builder = self.make_branch_builder('simple-branch')
-        builder.build_snapshot('A-id', None, [
-            ('add', ('', 'root-id', 'directory', None)),
-            ('add', ('file', 'file-id', 'file', 'content\n'))])
-        builder.build_snapshot('B-id', None, [])
-        builder.build_snapshot('C-id', None, [])
-        b = builder.get_branch()
-        if not b.repository._format.supports_chks:
-            raise TestNotApplicable('requires repository with chk_bytes')
-        b.lock_read()
-        self.addCleanup(b.unlock)
-        # check our setup: B-id and C-id should have identical chk root keys.
-        inv_b = b.repository.get_inventory('B-id')
-        inv_c = b.repository.get_inventory('C-id')
-        self.assertEqual(inv_b.id_to_entry.key(), inv_c.id_to_entry.key())
-        # Now, manually insert objects for a stacked repo with only revision
-        # C-id:
-        # We need ('revisions', 'C-id'), ('inventories', 'C-id'),
-        # ('inventories', 'B-id'), and the corresponding chk roots for those
-        # inventories.
-        repo = self.make_repository('damaged-repo')
-        repo.lock_write()
-        repo.start_write_group()
-        src_repo = b.repository
-        repo.inventories.insert_record_stream(
-            src_repo.inventories.get_record_stream(
-                [('B-id',), ('C-id',)], 'unordered', True))
-        repo.revisions.insert_record_stream(
-            src_repo.revisions.get_record_stream(
-                [('C-id',)], 'unordered', True))
-        # Make sure the presence of the missing data in a fallback does not
-        # avoid the error.
-        repo.add_fallback_repository(b.repository)
-        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
-        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
-        self.assertRaises(
-            errors.BzrCheckError, reopened_repo.commit_write_group)
-        reopened_repo.abort_write_group()
-
-    def test_missing_chk_root_for_parent_inventory(self):
-        """commit_write_group fails with BzrCheckError when the chk root record
-        for a parent inventory of a new revision is missing.
-        """
-        builder = self.make_branch_builder('simple-branch')
-        builder.build_snapshot('A-id', None, [
-            ('add', ('', 'root-id', 'directory', None)),
-            ('add', ('file', 'file-id', 'file', 'content\n'))])
-        builder.build_snapshot('B-id', None, [])
-        builder.build_snapshot('C-id', None, [
-            ('modify', ('file-id', 'new-content'))])
-        b = builder.get_branch()
-        if not b.repository._format.supports_chks:
-            raise TestNotApplicable('requires repository with chk_bytes')
-        b.lock_read()
-        self.addCleanup(b.unlock)
-        # Now, manually insert objects for a stacked repo with only revision
-        # C-id, *except* the chk root entry for the parent inventory.
-        # We need ('revisions', 'C-id'), ('inventories', 'C-id'),
-        # ('inventories', 'B-id'), and the corresponding chk roots for those
-        # inventories.
-        inv_c = b.repository.get_inventory('C-id')
-        chk_keys_for_c_only = [
-            inv_c.id_to_entry.key(), inv_c.parent_id_basename_to_file_id.key()]
-        repo = self.make_repository('damaged-repo')
-        repo.lock_write()
-        repo.start_write_group()
-        src_repo = b.repository
-        repo.chk_bytes.insert_record_stream(
-            src_repo.chk_bytes.get_record_stream(
-                chk_keys_for_c_only, 'unordered', True))
-        repo.inventories.insert_record_stream(
-            src_repo.inventories.get_record_stream(
-                [('B-id',), ('C-id',)], 'unordered', True))
-        repo.revisions.insert_record_stream(
-            src_repo.revisions.get_record_stream(
-                [('C-id',)], 'unordered', True))
-        # Make sure the presence of the missing data in a fallback does not
-        # avoid the error.
-        repo.add_fallback_repository(b.repository)
-        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
-        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
-        self.assertRaises(
-            errors.BzrCheckError, reopened_repo.commit_write_group)
-        reopened_repo.abort_write_group()
-
 
 class TestResumeableWriteGroup(TestCaseWithRepository):
 

=== modified file 'bzrlib/tests/per_repository_chk/__init__.py'
--- bzrlib/tests/per_repository_chk/__init__.py	2009-04-09 20:23:07 +0000
+++ bzrlib/tests/per_repository_chk/__init__.py	2009-09-08 06:25:26 +0000
@@ -46,8 +46,8 @@
 
 class TestCaseWithRepositoryCHK(TestCaseWithRepository):
 
-    def make_repository(self, path):
-        TestCaseWithRepository.make_repository(self, path)
+    def make_repository(self, path, format=None):
+        TestCaseWithRepository.make_repository(self, path, format=format)
         return repository.Repository.open(self.get_transport(path).base)
 
 

=== modified file 'bzrlib/tests/per_repository_chk/test_supported.py'
--- bzrlib/tests/per_repository_chk/test_supported.py	2009-04-09 20:23:07 +0000
+++ bzrlib/tests/per_repository_chk/test_supported.py	2009-09-08 06:25:26 +0000
@@ -16,7 +16,10 @@
 
 """Tests for repositories that support CHK indices."""
 
-from bzrlib import osutils
+from bzrlib import (
+    errors,
+    osutils,
+    )
 from bzrlib.versionedfile import VersionedFiles
 from bzrlib.tests.per_repository_chk import TestCaseWithRepositoryCHK
 
@@ -104,3 +107,244 @@
             self.assertEqual(expected_set, repo.chk_bytes.keys())
         finally:
             repo.unlock()
+
+
+class TestCommitWriteGroupIntegrityCheck(TestCaseWithRepositoryCHK):
+    """Tests that commit_write_group prevents various kinds of invalid data
+    from being committed to a CHK repository.
+    """
+
+    def reopen_repo_and_resume_write_group(self, repo):
+        resume_tokens = repo.suspend_write_group()
+        repo.unlock()
+        reopened_repo = repo.bzrdir.open_repository()
+        reopened_repo.lock_write()
+        self.addCleanup(reopened_repo.unlock)
+        reopened_repo.resume_write_group(resume_tokens)
+        return reopened_repo
+
+    def test_missing_chk_root_for_inventory(self):
+        """commit_write_group fails with BzrCheckError when the chk root record
+        for a new inventory is missing.
+        """
+        repo = self.make_repository('damaged-repo')
+        builder = self.make_branch_builder('simple-branch')
+        builder.build_snapshot('A-id', None, [
+            ('add', ('', 'root-id', 'directory', None)),
+            ('add', ('file', 'file-id', 'file', 'content\n'))])
+        b = builder.get_branch()
+        b.lock_read()
+        self.addCleanup(b.unlock)
+        repo.lock_write()
+        repo.start_write_group()
+        # Now, add the objects manually
+        text_keys = [('file-id', 'A-id'), ('root-id', 'A-id')]
+        # Directly add the texts, inventory, and revision object for 'A-id' --
+        # but don't add the chk_bytes.
+        src_repo = b.repository
+        repo.texts.insert_record_stream(src_repo.texts.get_record_stream(
+            text_keys, 'unordered', True))
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('A-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('A-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
+    def test_missing_chk_root_for_unchanged_inventory(self):
+        """commit_write_group fails with BzrCheckError when the chk root record
+        for a new inventory is missing, even if the parent inventory is present
+        and has identical content (i.e. the same chk root).
+        
+        A stacked repository containing only a revision with an identical
+        inventory to its parent will still have the chk root records for those
+        inventories.
+
+        (In principle the chk records are unnecessary in this case, but in
+        practice bzr 2.0rc1 (at least) expects to find them.)
+        """
+        repo = self.make_repository('damaged-repo')
+        # Make a branch where the last two revisions have identical
+        # inventories.
+        builder = self.make_branch_builder('simple-branch')
+        builder.build_snapshot('A-id', None, [
+            ('add', ('', 'root-id', 'directory', None)),
+            ('add', ('file', 'file-id', 'file', 'content\n'))])
+        builder.build_snapshot('B-id', None, [])
+        builder.build_snapshot('C-id', None, [])
+        b = builder.get_branch()
+        b.lock_read()
+        self.addCleanup(b.unlock)
+        # check our setup: B-id and C-id should have identical chk root keys.
+        inv_b = b.repository.get_inventory('B-id')
+        inv_c = b.repository.get_inventory('C-id')
+        self.assertEqual(inv_b.id_to_entry.key(), inv_c.id_to_entry.key())
+        # Now, manually insert objects for a stacked repo with only revision
+        # C-id:
+        # We need ('revisions', 'C-id'), ('inventories', 'C-id'),
+        # ('inventories', 'B-id'), and the corresponding chk roots for those
+        # inventories.
+        repo.lock_write()
+        repo.start_write_group()
+        src_repo = b.repository
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('B-id',), ('C-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('C-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
+    def test_missing_chk_leaf_for_inventory(self):
+        """commit_write_group fails with BzrCheckError when the chk root record
+        for a parent inventory of a new revision is missing.
+        """
+        repo = self.make_repository('damaged-repo')
+        b = self.make_branch_with_multiple_chk_nodes()
+        src_repo = b.repository
+        src_repo.lock_read()
+        self.addCleanup(src_repo.unlock)
+        # Now, manually insert objects for a stacked repo with only revision
+        # C-id, *except* drop the non-root chk records.
+        inv_b = src_repo.get_inventory('B-id')
+        inv_c = src_repo.get_inventory('C-id')
+        chk_root_keys_only = [
+            inv_b.id_to_entry.key(), inv_b.parent_id_basename_to_file_id.key(),
+            inv_c.id_to_entry.key(), inv_c.parent_id_basename_to_file_id.key()]
+        all_chks = src_repo.chk_bytes.keys()
+        # Pick a non-root key to drop
+        key_to_drop = all_chks.difference(chk_root_keys_only).pop()
+        all_chks.discard(key_to_drop)
+        repo.lock_write()
+        repo.start_write_group()
+        repo.chk_bytes.insert_record_stream(
+            src_repo.chk_bytes.get_record_stream(
+                all_chks, 'unordered', True))
+        repo.texts.insert_record_stream(
+            src_repo.texts.get_record_stream(
+                src_repo.texts.keys(), 'unordered', True))
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('B-id',), ('C-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('C-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
+    def test_missing_chk_root_for_parent_inventory(self):
+        """commit_write_group fails with BzrCheckError when the chk root record
+        for a parent inventory of a new revision is missing.
+        """
+        repo = self.make_repository('damaged-repo')
+        b = self.make_branch_with_multiple_chk_nodes()
+        b.lock_read()
+        self.addCleanup(b.unlock)
+        # Now, manually insert objects for a stacked repo with only revision
+        # C-id, *except* the chk root entry for the parent inventory.
+        # We need ('revisions', 'C-id'), ('inventories', 'C-id'),
+        # ('inventories', 'B-id'), and the corresponding chk roots for those
+        # inventories.
+        inv_c = b.repository.get_inventory('C-id')
+        chk_keys_for_c_only = [
+            inv_c.id_to_entry.key(), inv_c.parent_id_basename_to_file_id.key()]
+        repo.lock_write()
+        repo.start_write_group()
+        src_repo = b.repository
+        repo.chk_bytes.insert_record_stream(
+            src_repo.chk_bytes.get_record_stream(
+                chk_keys_for_c_only, 'unordered', True))
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('B-id',), ('C-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('C-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
+    def make_branch_with_multiple_chk_nodes(self):
+        # add and modify files with very long file-ids, so that the chk map
+        # will need more than just a root node.
+        builder = self.make_branch_builder('simple-branch')
+        file_adds = []
+        file_modifies = []
+        for char in 'abc':
+            name = char * 10000
+            file_adds.append(
+                ('add', ('file-' + name, 'file-%s-id' % name, 'file',
+                         'content %s\n' % name)))
+            file_modifies.append(
+                ('modify', ('file-%s-id' % name, 'new content %s\n' % name)))
+        builder.build_snapshot('A-id', None, [
+            ('add', ('', 'root-id', 'directory', None))] +
+            file_adds)
+        builder.build_snapshot('B-id', None, [])
+        builder.build_snapshot('C-id', None, file_modifies)
+        return builder.get_branch()
+        
+    def test_missing_text_record(self):
+        """commit_write_group fails with BzrCheckError when a text is missing.
+        """
+        repo = self.make_repository('damaged-repo')
+        b = self.make_branch_with_multiple_chk_nodes()
+        src_repo = b.repository
+        src_repo.lock_read()
+        self.addCleanup(src_repo.unlock)
+        # Now, manually insert objects for a stacked repo with only revision
+        # C-id, *except* drop one changed text.
+        all_texts = src_repo.texts.keys()
+        all_texts.remove(('file-%s-id' % ('c'*10000,), 'C-id'))
+        repo.lock_write()
+        repo.start_write_group()
+        repo.chk_bytes.insert_record_stream(
+            src_repo.chk_bytes.get_record_stream(
+                src_repo.chk_bytes.keys(), 'unordered', True))
+        repo.texts.insert_record_stream(
+            src_repo.texts.get_record_stream(
+                all_texts, 'unordered', True))
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('B-id',), ('C-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('C-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
+
+
+

=== modified file 'bzrlib/tests/per_repository_reference/test_add_revision.py'
--- bzrlib/tests/per_repository_reference/test_add_revision.py	2009-04-09 20:23:07 +0000
+++ bzrlib/tests/per_repository_reference/test_add_revision.py	2009-09-08 14:51:23 +0000
@@ -40,6 +40,7 @@
             repo.start_write_group()
             try:
                 rev = tree.branch.repository.get_revision(revid)
+                repo.texts.add_lines((inv.root.file_id, revid), [], [])
                 repo.add_revision(revid, rev, inv=inv)
             except:
                 repo.abort_write_group()

=== modified file 'bzrlib/tests/per_repository_reference/test_add_signature_text.py'
--- bzrlib/tests/per_repository_reference/test_add_signature_text.py	2009-04-09 20:23:07 +0000
+++ bzrlib/tests/per_repository_reference/test_add_signature_text.py	2009-09-08 14:51:23 +0000
@@ -39,13 +39,13 @@
             repo.start_write_group()
             try:
                 rev = tree.branch.repository.get_revision(revid)
+                repo.texts.add_lines((inv.root.file_id, revid), [], [])
                 repo.add_revision(revid, rev, inv=inv)
                 repo.add_signature_text(revid, "text")
+                repo.commit_write_group()
             except:
                 repo.abort_write_group()
                 raise
-            else:
-                repo.commit_write_group()
         finally:
             repo.unlock()
         repo.get_signature_text(revid)

=== modified file 'bzrlib/tests/per_workingtree/test_parents.py'
--- bzrlib/tests/per_workingtree/test_parents.py	2009-07-10 07:14:02 +0000
+++ bzrlib/tests/per_workingtree/test_parents.py	2009-09-09 01:05:31 +0000
@@ -392,6 +392,14 @@
             try:
                 if shape.root.revision is None:
                     shape.root.revision = revid
+                # Create the text records for this inventory.
+                for path, ie in shape.iter_entries():
+                    if ie.text_size:
+                        lines = ['a' * ie.text_size]
+                    else:
+                        lines = []
+                    tree.branch.repository.texts.add_lines(
+                        (ie.file_id, ie.revision), [], lines)
                 sha1 = tree.branch.repository.add_inventory(revid, shape, [])
                 rev = Revision(timestamp=0,
                                timezone=None,
@@ -400,11 +408,10 @@
                                inventory_sha1=sha1,
                                revision_id=revid)
                 tree.branch.repository.add_revision(revid, rev)
+                tree.branch.repository.commit_write_group()
             except:
                 tree.branch.repository.abort_write_group()
                 raise
-            else:
-                tree.branch.repository.commit_write_group()
         finally:
             tree.unlock()
 

=== modified file 'bzrlib/tests/test_inv.py'
--- bzrlib/tests/test_inv.py	2009-07-20 05:19:53 +0000
+++ bzrlib/tests/test_inv.py	2009-09-09 01:28:34 +0000
@@ -65,6 +65,14 @@
     return multiply_tests(to_adapt, scenarios, result)
 
 
+def create_texts_for_inv(repo, inv):
+    for path, ie in inv.iter_entries():
+        if ie.text_size:
+            lines = ['a' * ie.text_size]
+        else:
+            lines = []
+        repo.texts.add_lines((ie.file_id, ie.revision), [], lines)
+    
 def apply_inventory_Inventory(self, basis, delta):
     """Apply delta to basis and return the result.
     
@@ -137,6 +145,7 @@
             rev = revision.Revision('basis', timestamp=0, timezone=None,
                 message="", committer="foo@example.com")
             basis.revision_id = 'basis'
+            create_texts_for_inv(tree.branch.repository, basis)
             repo.add_revision('basis', rev, basis)
             # Add a revision for the result, with the basis content - 
             # update_basis_by_delta doesn't check that the delta results in
@@ -146,11 +155,10 @@
                 message="", committer="foo@example.com")
             basis.revision_id = 'result'
             repo.add_revision('result', rev, basis)
+            repo.commit_write_group()
         except:
             repo.abort_write_group()
             raise
-        else:
-            repo.commit_write_group()
         # Set the basis state as the trees current state
         tree._write_inventory(basis)
         # This reads basis from the repo and puts it into the tree's local
@@ -221,12 +229,12 @@
             rev = revision.Revision('basis', timestamp=0, timezone=None,
                 message="", committer="foo@example.com")
             basis.revision_id = 'basis'
+            create_texts_for_inv(repo, basis)
             repo.add_revision('basis', rev, basis)
+            repo.commit_write_group()
         except:
             repo.abort_write_group()
             raise
-        else:
-            repo.commit_write_group()
     finally:
         repo.unlock()
     repo.lock_write()

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2009-08-27 05:22:14 +0000
+++ bzrlib/tests/test_remote.py	2009-09-09 01:38:48 +0000
@@ -2326,6 +2326,10 @@
         class FakeRealRepository:
             def _get_sink(self):
                 return fake_real_sink
+            def is_in_write_group(self):
+                return False
+            def refresh_data(self):
+                return True
         repo._real_repository = FakeRealRepository()
         sink = repo._get_sink()
         fmt = repository.RepositoryFormat.get_default_format()

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2009-09-04 03:58:41 +0000
+++ bzrlib/tests/test_repository.py	2009-09-09 01:58:47 +0000
@@ -976,6 +976,7 @@
             inv = inventory.Inventory(revision_id='rev1a')
             inv.root.revision = 'rev1a'
             self.add_file(repo, inv, 'file1', 'rev1a', [])
+            repo.texts.add_lines((inv.root.file_id, 'rev1a'), [], [])
             repo.add_inventory('rev1a', inv, [])
             revision = _mod_revision.Revision('rev1a',
                 committer='jrandom@example.com', timestamp=0,
@@ -1016,6 +1017,7 @@
     def add_revision(self, repo, revision_id, inv, parent_ids):
         inv.revision_id = revision_id
         inv.root.revision = revision_id
+        repo.texts.add_lines((inv.root.file_id, revision_id), [], [])
         repo.add_inventory(revision_id, inv, parent_ids)
         revision = _mod_revision.Revision(revision_id,
             committer='jrandom@example.com', timestamp=0, inventory_sha1='',

